"""
Parse .xml or .mxl files and return a pandas DataFrameself.
"""
import sys
import os

import numpy as np
import pandas as pd

# import xml parser from magenta
from pitchplots.modified_musicxml_parser import MusicXMLDocument

class ParseError(Exception):
    """
    Exception thrown when the MusicXML contents cannot be parsed.
    """
    pass

### DEFINE PARSER
def xml_to_csv(filepath=os.path.dirname(os.path.realpath(__file__))+'\\'+'data'+'\\'+'data_example.mxl',
               filename=None, save_csv=True, duration='whole_note'):
    """return the Dataframe, and possbily register it in csv, of the musicxml file
    
    Keyword arguments:
    filepath -- absolute path to the xml file by default goes to the example file
    filename -- give the name of the .csv file, by default give the same name as the .mxl file
    save_cvs -- if True save the csv file in the csv directory or at the given path
    duration -- define of the duration will be in seconds or relative to a whole note
                (possible values: 'seconds' or 'whole_note'(default value))
    """
    columns = ['filepath', # piece ID or something (TODO)
               'qpm', #add qpm, the beat per minute
               'time_sign_num', #add the time signature numerator
               'time_sign_den', #add the time signature denumerator
               'measure_no', # number of measure
               'no_accs', # number of accidentals
               'mode', # mode of key as defined in XML (not reliable)
               'key_area', # begin of key signature
               'type', # type of event (note or rest)
               'note_name', # note name (e.g. C4, Bb2)
               'tpc', # tonal pitch class: note name w/o octave, e.g. C, F#...
               'step', # diatonic step in A, B, C, D, E, F, G
               'acc', # accidentals
               'octave', # octave number (int)
               'pitch', # MIDI pitch name (i.e. C4 = 60)
               'pitch_class', # pitch modulo 12
               'duration', # note duration in beats as float (i.e. a quarter note is 0.25) ???? IS THIS CORRECT?
               'onset' # onset in seconds
              ]
    
    #add of these variables for control
    key_signature_on = False
    time_signature_on = False
    qpm_on = False
    
    try:
        parsed = MusicXMLDocument(filepath)
    except:
        raise ParseError('There is a problem with the path to the xml/mxl file or the files are not standard.')
    
    df = pd.DataFrame(columns=columns)

    for part in parsed.parts:
        measure_no = 0
        for measure in part.measures:
            measure_no += 1
            
            #keep the previous key signature
            #because the key signature appears only if it changes
            if pd.isnull(measure.key_signature) == False:
                key_signature_on = True
                root = measure.key_signature.key
                mode = measure.key_signature.mode
                key_area = measure.key_signature.time_position
            elif key_signature_on == False:
                root = np.nan
                mode = np.nan
                key_area = np.nan
            
            #adding of the time_signature like the key_signature
            if pd.isnull(measure.time_signature) == False:
                time_signature_on = True
                time_sign_num = measure.time_signature.numerator
                time_sign_den = measure.time_signature.denominator
            elif time_signature_on == False:
                time_sign_num = np.nan
                time_sign_den = np.nan
            
            #adding of qpm
            if pd.isnull(measure.state.qpm) == False:
                qpm_on = True
                qpm = measure.state.qpm
            elif qpm_on == False:
                qpm = np.nan
            
            for note in measure.notes:
                if note.is_rest:
                    ntype = 'rest'
                else:
                    ntype = 'note'
                if note.pitch is not None:
                    note_name = note.pitch[0]
                    tpc = note_name[:-1]
                    pitch = int(note.pitch[1])
                    pitch_class = int(note.pitch[1] % 12)
                    step = note_name[0]
                    if '#' in note_name:
                        acc = 1
                    elif 'bb' in note_name:
                        acc = -2
                    elif 'b' in note_name:
                        acc = -1
                    elif 'x' in note_name:
                        acc = 2
                    else:
                        acc = 0
                    octave = note_name[-1]
                else:
                    note_name = np.nan
                    tpc = np.nan
                    pitch = np.nan
                    step = np.nan
                    acc = np.nan
                    pitch_class = np.nan
                    octave = np.nan
                duration =  note.note_duration.duration_float()
                onset = note.note_duration.time_position

                # final list of of the columns of the dataframe
                values = [filepath,
                          qpm,
                          time_sign_num,
                          time_sign_den,
                          measure_no,
                          root,
                          mode,
                          key_area,
                          ntype,
                          note_name,
                          tpc,
                          step,
                          acc,
                          octave,
                          pitch,
                          pitch_class,
                          duration,
                          onset]
                row = dict(zip(columns, values))
                df = df.append(row, ignore_index=True)
    
    # correct the onset to be quantized by the measure number
    # add the 'onset_seconds' column from the new onset, for the dynamic plotting
    df = data_onset_duration_corrector(df, duration)
    
    if save_csv:
        #  path to the csv directory
        csv_path = os.path.dirname(sys.argv[0])+r'/csv'
        # get the name from the xml file and put in csv dir
        if pd.isnull(filename):
            if not os.path.exists(csv_path):
                os.makedirs(csv_path)
            filename = os.path.basename(filepath).split('.')[0] + '.csv'
            df.to_csv(os.path.join(csv_path,filename), sep=',')
        # if filename is a path, register the csv file at the given path
        elif "\\" in filename or "/" in filename:
            df.to_csv(filename, sep=',')
        # get the name and put it in csv folder
        else:
            # check if the csv folder already exist if not create one
            if not os.path.exists(csv_path):
                os.makedirs(csv_path)
            df.to_csv(os.path.join(csv_path,filename), sep=',')

    return df

def data_onset_duration_corrector(data, duration):
    """
    corrects the duration and onset of the piece and normalize by the measure
    
    Keyword arguments:
    data -- the pandas DataFrame of the piece
    duration -- define of the duration will be in seconds or relative to a whole note
                (possible values: 'seconds' or 'whole_note'(default value))
    return:
    ret_data -- the pandas DataFrame of the the correction of the piece
    """
    corr_data = data.copy()
    
    #get rid of notes and rests that have a 0 duration
    #they are more likely falsely parsed notes from musicxml_parser
    corr_data.drop(corr_data[corr_data.duration == 0].index, inplace=True)
    
    min_onset = 0
    max_onset = 0
    onset_ratio = 0
    duration_ratio = 0
    current_minimum_time = 0
    time_ratio = 0
    
    #get rid of all the notes that have a zero duration
    corr_data.drop(corr_data[corr_data.duration == 0].index, inplace=True)
    
    #group the notes by measure 
    gb_mesure_no = corr_data.groupby('measure_no')
    
    ret_data = pd.DataFrame()
    
    for i in range(corr_data['measure_no'].max()):
        df_group = gb_mesure_no.get_group(i+1).copy()
        
        #assume that the onset_ratio is the same for the last and before last measure
        if i != corr_data['measure_no'].max() - 1:
            df_group_next = gb_mesure_no.get_group(i+1+1)
            
            #it is assumed that the first note of the measure begins at the start of the measure
            min_onset = df_group['onset'].min()
            max_onset = df_group_next['onset'].min()
            onset_ratio = 1/(max_onset-min_onset)
        
        #set the onset like the first note start at the at the start of the onset
        df_group['onset'] *= onset_ratio
        df_group['onset'] += (i - df_group['onset'].min())
        
        #the time signature is the same for the whole measure
        duration_ratio = df_group['time_sign_den'].iloc[0]/df_group['time_sign_num'].iloc[0]
        
        # if True change the duration to be in seconds using the BPM value
        if duration =='seconds':
            #so the duration is equal to the number of seconds of the quatized note
            df_group['duration'] *= (4*60)/df_group['qpm'].iloc[0]
        
        #the ratio between the time in second when the note is played and the measure relative timing
        time_ratio = (4 * 60)/(duration_ratio * df_group['qpm'].iloc[0])
        
        #the adding of the time column using the onset column as base
        df_group = df_group.assign(onset_seconds=df_group['onset'].values)
        df_group['onset_seconds'] += -i
        df_group['onset_seconds'] *= time_ratio
        df_group['onset_seconds'] += current_minimum_time
        
        #the calculation of the time in second when the next measure begins
        current_minimum_time = current_minimum_time + time_ratio
        
        #add the group to the note sequence
        ret_data = pd.concat([ret_data, df_group])
        
    return ret_data


"""
stack of functions for the program
"""
import numpy as np
import pandas as pd

def get_dic_nei(pitch_class_display):
    """for musical_plot_hex it need the neighbouring notes in the hexagonal shape, return dictionary
        ref is the list of note that we are refering to
        pos is the position of the neighbouring note that we want to know
        note is the note at the position from the refering note
        sup is the accidental of the note at the given position
    """
    if pitch_class_display:
        # for every note there is a row 
        dic_nei = {'ref':[i for sublist in [[i]*6 for i in range(12)] for i in sublist],
                 'pos': [(1,0,-1), (1,-1,0), (0,-1,1), (-1,0,1), (-1,1,0), (0,1,-1)]*12,
                 'note': [4, 7, 3, 8, 5, 9,
                          5, 8, 4, 9, 6, 10,
                          6, 9, 5, 10, 7, 11,
                          7, 10, 6, 11, 8, 0,
                          8, 11, 7, 0, 9, 1,
                          9, 0, 8, 1, 10, 2,
                          10, 1, 9, 2, 11, 3,
                          11, 2, 10, 3, 0, 4,
                          0, 3, 11, 4, 1, 5,
                          1, 4, 0, 5, 2, 6,
                          2, 5, 1, 6, 3, 7,
                          3, 6, 2, 7, 4, 8]}
    else:
        dic_nei = {'ref':[c for sublist in [[c]*6 for c in 'FCGDAEB'] for c in sublist],
                 'pos': [(1,0,-1), (1,-1,0), (0,-1,1), (-1,0,1), (-1,1,0), (0,1,-1)]*7,
                 'note': ['A', 'C', 'A', 'D', 'B', 'D',
                          'E', 'G', 'E', 'A', 'F', 'A',
                          'B', 'D', 'B', 'E', 'C', 'E',
                          'F', 'A', 'F', 'B', 'G', 'B',
                          'C', 'E', 'C', 'F', 'D', 'F',
                          'G', 'B', 'G', 'C', 'A', 'C',
                          'D', 'F', 'D', 'G', 'E', 'G'],
                 'acc': [0, 0, -1, -1, -1, 0,
                         0, 0, -1, -1, 0, 0,
                         0, 0, -1, -1, 0, 0,
                         1, 0, 0, -1, 0, 0,
                         1, 0, 0, 0, 0, 1,
                         1, 0, 0, 0, 0, 1,
                         1, 1, 0, 0, 0, 1]}
    return dic_nei

def sampling(value, sampling_frequency):
    """return the sampled value at a given sampling frequency"""
    ret_value = 0
    
    if value%(1/sampling_frequency) <= (1/sampling_frequency)/2:
        ret_value = value - (value%(1/sampling_frequency))
    else:
        ret_value = value - (value%(1/sampling_frequency)) + (1/sampling_frequency)
        
    return ret_value

def put_flat_sharp(step, acc):
    """get a step and its acc and return the note in tpc notation, return str"""
    ret_note = step # not needed
    ret_acc = acc
    
    #put the flats and sharps
    if ret_acc > 0:
        for l in range(ret_acc):
            ret_note = ret_note + '#'     
    if ret_acc < 0:
        ret_acc = abs(ret_acc)
        for l in range(ret_acc):
            ret_note = ret_note + 'b'
    return ret_note # not needed

def is_tpc(note):
    """check if note has the same format as tpc, return boolean"""
    s_tpc_values_1 = pd.Series(['F', 'C', 'G', 'D', 'A', 'E', 'B'])
    s_tpc_values_2 = pd.Series(['b', '#'])

    correct_format_tpc = True
    count = 0

    for i in str(note):
        count = count + 1

        #check first character
        if (count == 1): 
            if (i == s_tpc_values_1).any() == False:
                correct_format_tpc = False
                    
        #check the flats and sharps
        else: 
            if (i == s_tpc_values_2).any() == False:
                correct_format_tpc = False
        if correct_format_tpc == False:
            break
    return correct_format_tpc

def is_pc(note):
    """check if note has the same format as pc, return boolean"""
    return note in range(12)

def get_acc(note):
    """get the acc from a tpc format, return int"""
    if is_tpc(note):
        acc = 0
        for i in str(note):
            if i == '#':
                acc = acc + 1
            if i == 'b':
                acc = acc - 1
        return acc

def get_step(note):
    """get the step from a tpc format, return str"""
    if is_tpc(note):
        step = ''
        step = str(note)[0]
        return step

def get_pc(note):
    """get the pitch class from a tpc value, return int"""
    pc = np.NaN
    s_tpc_pc = pd.Series(data=[0, 2, 4, 5, 7, 9, 11], index=['C', 'D', 'E', 'F', 'G', 'A', 'B'])
    if pd.isnull(note) == False:
        pc = s_tpc_pc.at[get_step(note)]
        pc = pc + get_acc(note)
        ###heck modulo
        while pc < 0:
            pc = pc + 12
        while pc > 11:
            pc = pc - 12
    return pc

def get_fifth_nb(note):
    """return the position of the note in the fifth line"""
    step = get_step(note)
    acc = get_acc(note)
    dic = {'F':0, 'C':1, 'G':2, 'D':3, 'A':4, 'E':5, 'B':6}
    return dic[step] + acc * 7

#change note to fifth_number Use same dictionnary
def get_fifth_note(note):
    """return the note after the fifth number"""
    count = 0
    ret_note = 0
    copy_note = int(note)
    while copy_note < 0 or copy_note > 6:
        if copy_note < 0:
            copy_note = copy_note + 7
            count = count - 1
        if copy_note > 6:
            copy_note = copy_note - 7
            count = count + 1
    if copy_note == 0: ret_note = 'F'
    if copy_note == 1: ret_note = 'C'
    if copy_note == 2: ret_note = 'G'
    if copy_note == 3: ret_note = 'D'
    if copy_note == 4: ret_note = 'A'
    if copy_note == 5: ret_note = 'E'
    if copy_note == 6: ret_note = 'B'
    ret_note = put_flat_sharp(ret_note, count)
    return ret_note



"""
module that has to read what is given to the plotting functions
"""
from pitchplots.functions import get_acc, get_step, get_pc, sampling

import pandas as pd
#from midiutil import MIDIFile
#import librosa
#import numpy

def get_df_short(
    piece,
    vocabulary={0:'C', 1:'Db', 2:'D', 3:'Eb', 4:'E', 5:'F', 6:'Gb', 7:'G', 8:'Ab', 9:'A', 10:'Bb', 11:'B'},
    pitch_type='tpc',
    duration=False,
    measures=None):
    """return a Dataframe with the condenced information about the piece for static plots

    Function:
    Get a file columns and read it to create a dataFrame with the following informations:
        pitch class value('pc'), number of appearences('nb'), total duration of the note('duration'),
        tpc format note('tpc'), sharps and flats('acc')
    Keyword arguments:
    piece -- the absolute path to the .csv file containing the data
    vocabulary -- the conversion table from pitch class to tpc(F#, A, ...) format,
        the position indicate the pitch class value (default [C, Db, D, Eb, E, F, Gb, G, Ab, A, Bb, B])
    pitch_type -- the type of data that contains the file (default 'tpc')
        (tpc:[A, B#, Gbbb, ...], pc (pitch class):[0, 3, 7, ...])
    duration -- tell him if he has to class the notes by their total duration or their number of appearance
    measures -- give a set of measures example [5, 18], will display the notes of the measures 5 to 18 included
    """ 
    #check if it is a path to .csv or a DataFrame
    if isinstance(piece, list):
        df_tpc = pd.DataFrame()
        df_tpc['tpc'] = piece
        df_tpc['nb'] = 1
        #get the pc values from tpc
        df_tpc['pc'] = df_tpc['tpc'].apply(get_pc)
        #add the sup column
        df_tpc['acc'] = df_tpc['tpc'].apply(get_acc)
        #add the step column
        df_tpc['step'] = df_tpc['tpc'].apply(get_step)
        return df_tpc
    
    if isinstance(piece, pd.DataFrame):
        df_data = piece.copy()
    else:
        df_data =  pd.read_csv(piece)
    
    #the column with the pc values is called pitch_class so it rename it to 'pc'
    if 'pitch_class' in df_data.columns:
        df_data.rename(columns={'pitch_class': 'pc'}, inplace=True)
        
    if type(measures) is list:
        df_data.drop(df_data[df_data.measure_no < measures[0]].index, inplace=True)
        df_data.drop(df_data[df_data.measure_no > measures[1]].index, inplace=True)
        df_data.reset_index(drop=True, inplace=True)
    
    #If pitch_type is "tpc" we get a small DataFrame with the the columns:
    #   tpc, nb, duration (if exist), pc, acc, step
    if 'tpc' in df_data.columns and pitch_type == 'tpc':
        s_tpc = df_data['tpc'].groupby(df_data['tpc']).size()
        if not duration:
            s_tpc.sort_values(inplace=True, ascending=False)
        df_tpc = pd.DataFrame({'tpc':s_tpc.index, 'nb':s_tpc.values})
        if 'duration' in df_data.columns:
            df_duration = pd.DataFrame(df_data.groupby('tpc').duration.sum())
            df_tpc = pd.merge(df_tpc, df_duration, on='tpc')
            if duration:
                df_tpc.sort_values('duration', inplace=True, ascending=False)
            
        #get the pc values from tpc
        df_tpc['pc'] = df_tpc['tpc'].apply(get_pc)
        #add the sup column
        df_tpc['acc'] = df_tpc['tpc'].apply(get_acc)
        #add the step column
        df_tpc['step'] = df_tpc['tpc'].apply(get_step)
        
        return df_tpc
    
    #If pitch_type is "pc" we get a small DataFrame with the the columns:
    #   pc, nb, duration (if exist), tpc(from vocabulary), acc, step
    if 'pc' in df_data.columns and pitch_type == 'pc':
        s_pc = df_data['pc'].groupby(df_data['pc']).size()
        if not duration:
            s_pc.sort_values(inplace=True, ascending=False)
        df_pc = pd.DataFrame({'pc':s_pc.index, 'nb':s_pc.values})

        #add the duration column
        if 'duration' in df_data.columns:
            df_duration = pd.DataFrame(df_data.groupby('pc').duration.sum())
            df_pc = pd.merge(df_pc, df_duration, on='pc')
            if duration:
                df_pc.sort_values('duration', inplace=True, ascending=False)
            
        #add the tpc column
        s_tpc_pc = []
        for i in range(df_pc.shape[0]):
            s_tpc_pc.append(vocabulary[int(df_pc.at[i, 'pc'])])

        df_pc['tpc'] = pd.Series(s_tpc_pc)

        #add the sup column
        df_pc['acc'] = df_pc['tpc'].apply(get_acc)

        #keep only the step from tpc
        df_pc['step'] = df_pc['tpc'].apply(get_step)

        return df_pc
    